module jp.ku6.bitmap.png;

public import jp.ku6.bitmap.core;

import etc.c.zlib;
import std.stream;

/++
 + PNGアクセスクラス
 +
 + 実装済み機能
 + ・ファイルからARGBピクセルデータとしてロード
 +
 + 対応フォーマット
 + ・インデックスカラー
 + ・RGBカラー
 + ・ARGBカラー
 +
 + 対応チャンク
 + ・基本チャンク (IHDR, IDAT, IEND)
 + ・パレットチャンク (PLTE)
 + ・透過色チャンク (tRNS)
 +/
class PortableNetworkGraphicsFile : Bitmap
{
	/++
	 + コンストラクタ
	 +
	 + Params:
	 +  filePath = ファイルパス
	 +/
	this(string filePath)
	{
		filePath_ = filePath;
	}

	@property
	{
		uint width()
		{
			return width_;
		}

		uint height()
		{
			return height_;
		}
	}

	bool open()
	{
		// ファイルをオープン
		fileStream_ = new BufferedFile(filePath_);

		// ヘッダ情報を読み出し
		ubyte[33] buffer;
		fileStream_.read(buffer);

		// 識別子を確認
		for (uint i; i < IDENTIFIER.length; ++i)
			if (buffer[i] != IDENTIFIER[i])
				goto OPEN_FAILED;

		// データ長を取得
		uint dataLength = (buffer[8] << 24) | (buffer[9] << 16) | (buffer[10] << 8) | buffer[11];

		// 名称を取得
		if (buffer[12] != cast(ubyte)'I' || buffer[13] != cast(ubyte)'H' || buffer[14] != cast(ubyte)'D' || buffer[15] != cast(ubyte)'R')
			goto OPEN_FAILED;

		// 横幅を取得
		width_ = (buffer[16] << 24) | (buffer[17] << 16) | (buffer[18] << 8) | buffer[19];

		// 高さを取得
		height_ = (buffer[20] << 24) | (buffer[21] << 16) | (buffer[22] << 8) | buffer[23];

		// ビット深度を取得
		depth_ = buffer[24];
		if (depth_ != 8)
			goto OPEN_FAILED;

		// カラー形式を取得
		switch (buffer[25])
		{
		case ColorType.GLAY_SCALE:
			goto OPEN_FAILED;
		case ColorType.RGB:
			colorType_ = ColorType.RGB;
			break;
		case ColorType.PALETTE:
			colorType_ = ColorType.PALETTE;
			break;
		case ColorType.GLAY_SCALE_WITH_ALPHA:
			goto OPEN_FAILED;
		case ColorType.RGB_WITH_ALPHA:
			colorType_ = ColorType.RGB_WITH_ALPHA;
			break;
		default:
			goto OPEN_FAILED;
		}

		// 圧縮形式を取得
		switch (buffer[26])
		{
		case CompressType.DEFAULT:
			break;
		default:
			goto OPEN_FAILED;
		}

		// フィルタ形式を取得
		switch (buffer[27])
		{
		case FilterType.DEFAULT:
			break;
		default:
			goto OPEN_FAILED;
		}

		// インターレース形式を取得
		switch (buffer[28])
		{
		case InterlaceType.NONE:
			break;
		case InterlaceType.ADAM7:
			goto OPEN_FAILED;
		default:
			goto OPEN_FAILED;
		}

		return true;

// 失敗時にはクローズしてから終わる
OPEN_FAILED:

		fileStream_.close();

		return false;
	}

	void close()
	{
		fileStream_.close();
	}

	uint[] readAsTrueColor()
	{
		uint[] destination = new uint[width_ * height_];
		return readAsTrueColor(destination, width_, height_);
	}

	uint[] readAsTrueColor(uint[] destination, uint destinationWidth, uint destinationHeight)
	{
		// 展開先バッファに展開できるサイズがあるか確認
		if (destination is null || destination.length == 0 || destinationWidth == 0 || destinationHeight == 0 || destination.length < destinationWidth * destinationHeight || destinationWidth < width_ || destinationHeight < height_)
			return null;

		// ファイルから変換元データ等をロード
		ubyte[] source = loadAndDecode();
		if (source is null)
			return null;

		// 1ピクセルのバイト数を取得
		uint bytePerPixel = getBytePerPixel();

		// 変換元データの横一列のバイト数を計算
		uint pitch = 1 + (width_ + 1) * bytePerPixel;

		// Full color
		if (colorType_ == ColorType.RGB)
		{
			if (depth_ == 8)
			{
				ubyte* readCursor = &source[pitch + 1 + bytePerPixel];
				uint* writeCursor = destination.ptr;

				if (transparentColor_ !is null)
				{
					uint transparentColor = 0xff000000 | transparentColor_[0] | (transparentColor_[1] << 8) | (transparentColor_[2] << 16);
					for (uint y; y < height_; ++y)
					{
						for (uint x; x < width_; ++x)
						{
							uint color = 0xff000000 | readCursor[x * 3] | (readCursor[x * 3 + 1] << 8) | (readCursor[x * 3 + 2] << 16);
							if (color == transparentColor)
								writeCursor[x] = 0;
							else
								writeCursor[x] = color;
						}
						readCursor += pitch;
						writeCursor += destinationWidth;
					}
				}
				else
				{
					for (uint y; y < height_; ++y)
					{
						for (uint x; x < width_; ++x)
							writeCursor[x] = 0xff000000 | readCursor[x * 3] | (readCursor[x * 3 + 1] << 8) | (readCursor[x * 3 + 2] << 16);
						readCursor += pitch;
						writeCursor += destinationWidth;
					}
				}
			}
		}
		// Index color
		else if (colorType_ == ColorType.PALETTE)
		{
			ubyte* readCursor = &source[pitch + 1 + bytePerPixel];
			uint* writeCursor = destination.ptr;

			if (transparentColor_ !is null)
			{
				ubyte transparentColor = transparentColor_[0];

				for (uint y; y < height_; ++y)
				{
					for (uint x; x < width_; ++x)
					{
						ubyte color = readCursor[x];
						if (color == transparentColor)
							writeCursor[x] = 0;
						else
							writeCursor[x] = palette_[color];
					}
					readCursor += pitch;
					writeCursor += destinationWidth;
				}
			}
			else
			{
				for (uint y; y < height_; ++y)
				{
					for (uint x; x < width_; ++x)
						*writeCursor++ = palette_[readCursor[x]];
					readCursor += pitch;
					writeCursor += destinationWidth;
				}
			}
		}
		// True color
		else if (colorType_ == ColorType.RGB_WITH_ALPHA)
		{
			if (depth_ == 8)
			{
				ubyte* readCursor = &source[pitch + 1 + bytePerPixel];
				uint* writeCursor = destination.ptr;
				for (uint y; y < height_; ++y)
				{
					for (uint x; x < width_; ++x)
					{
						uint color;
						color = readCursor[x * 4];
						color |= readCursor[x * 4 + 1] << 8;
						color |= readCursor[x * 4 + 2] << 16;
						color |= readCursor[x * 4 + 3] << 24;
						writeCursor[x] = color;
					}
					readCursor += pitch;
					writeCursor += destinationWidth;
				}
			}
		}

		return destination;
	}

private:

	// 識別子
	immutable static
	ubyte[8] IDENTIFIER = [0x89, 0x50, 0x4e, 0x47, 0x0d, 0x0a, 0x1a, 0x0a];

	// 色形式
	enum ColorType : ubyte
	{
		GLAY_SCALE = 0,
		RGB = 2,
		PALETTE = 3,
		GLAY_SCALE_WITH_ALPHA = 4,
		RGB_WITH_ALPHA = 6,
	}

	// 圧縮形式
	enum CompressType : ubyte
	{
		DEFAULT = 0,
	}

	// フィルタ形式
	enum FilterType : ubyte
	{
		DEFAULT = 0,
	}

	// インターレース形式
	enum InterlaceType : ubyte
	{
		NONE = 0,
		ADAM7 = 1,
	}

	// ファイルパス
	string filePath_;

	// ファイルストリーム
	BufferedFile fileStream_;

	// 画像の幅
	uint width_;

	// 画像の高さ
	uint height_;

	// 画像のビット深度 1ピクセル辺りではないことに注意
	ubyte depth_;

	// 画像の色形式
	ColorType colorType_;

	// 透明色データ
	ubyte[] transparentColor_;

	// パレットデータ
	uint[] palette_;

	/+
	 + 1ピクセルのバイト数を取得
	 +
	 + Returns:
	 +  バイト数
	 +/
	const
	uint getBytePerPixel()
	{
		uint bytePerPixel = depth_ / 8;
		if (colorType_ == ColorType.RGB)
			bytePerPixel *= 3;
		else if (colorType_ == ColorType.GLAY_SCALE_WITH_ALPHA)
			bytePerPixel *= 2;
		else if (colorType_ == ColorType.RGB_WITH_ALPHA)
			bytePerPixel *= 4;
		return bytePerPixel;
	}

	/++
	 + ファイルからデータをロード
	 +
	 + ファイルから必要なデータをロードする
	 + また、データチャンクについては解凍とフィルタの解除を行う
	 +
	 + Returns:
	 +  デコードされたデータチャンクのデータ
	 +  失敗時にはnullを返す
	 +/
	ubyte[] loadAndDecode()
	{
		// IHDRチャンクの直後から読み込みを開始
		fileStream_.seek(33, SeekPos.Set);

		// 1ピクセルのバイト数を取得
		uint bytePerPixel = getBytePerPixel();

		// フィルタバイトを含めたデータが解凍出来る領域を確保する
		//
		// フィルタのデコードで、左、左上、上のピクセルを参照するので、
		// 左と上に1ピクセルずつダミーデータを挿入する
		//
		// つまり、次のような構造になる
		//
		// [ダミーフィルタバイト][ダミーピクセル][ダミーピクセルデータ]
		// [   フィルタバイト   ][ダミーピクセル][  実ピクセルデータ  ]
		uint pitch = 1 + (width_ + 1) * bytePerPixel;
		ubyte[] buffer = new ubyte[pitch * (height_ + 1)];

		// ストリーム初期化
		etc.c.zlib.z_stream zstream;
	    if (etc.c.zlib.inflateInit2(&zstream, 15) != Z_OK)
			return null;

		// 出力領域を設定
		// バッファの後ろ側に詰める
		zstream.next_out = &buffer[pitch + bytePerPixel * height_];
		zstream.avail_out = buffer.length - (pitch + bytePerPixel * height_);

		// 解凍完了フラグ
		bool inflateComplete = false;

		// チャンクデータ読み込み領域
		ubyte[] chunkData;

		// 全チャンネルを調べ、IDATチャンクのみを処理する
		while (!fileStream_.eof)
		{
			// データ長と名前を読み込み
			ubyte[8] header;
			fileStream_.read(header);

			// データ長
			uint dataLength = (header[0] << 24) | (header[1] << 16) | (header[2] << 8) | header[3];

			// 名前
			uint name = (header[4] << 24) | (header[5] << 16) | (header[6] << 8) | header[7];

			// IDAT
			if (name == 0x49444154)
			{
				// チャンクデータ読み込み領域を確保
				if (chunkData.length < dataLength)
					chunkData.length = dataLength;

				// チャンクデータを読み込み
				fileStream_.read(chunkData[0 .. dataLength]);

				// チャンクのデータを入力データとして入力にセット
				zstream.next_in = cast(ubyte*)chunkData.ptr;
				zstream.avail_in = dataLength;

				// 入力データを全て処理するまでループ
				while (zstream.avail_in > 0)
				{
					int inflateResult = etc.c.zlib.inflate(&zstream, Z_NO_FLUSH);
					if (inflateResult == Z_STREAM_END)
					{
						if (etc.c.zlib.inflateEnd(&zstream) != Z_OK)
							return null;
						inflateComplete = true;
						break;
					}
					else if (inflateResult != Z_OK)
					{
						etc.c.zlib.inflateEnd(&zstream);
						return null;
					}
				}
			}
			// tRNS
			else if (name == 0x74524E53)
			{
				transparentColor_ = new ubyte[dataLength];
				fileStream_.read(transparentColor_);
			}
			// PLTE
			else if (name == 0x504C5445)
			{
				// チャンクデータ読み込み領域を確保
				if (chunkData.length < dataLength)
					chunkData.length = dataLength;

				// チャンクデータを読み込み
				fileStream_.read(chunkData[0 .. dataLength]);

				// パレットを作成
				uint colors = dataLength / 3;
				palette_ = new uint[colors];
				for (uint i; i < colors; ++i)
					palette_[i] = 0xff000000 | chunkData[i * 3] | (chunkData[i * 3 + 1] << 8) | (chunkData[i * 3 + 2] << 16);
			}
			// IEND
			else if (name == 0x49454E44)
			{
				if (!inflateComplete)
				{
					etc.c.zlib.inflateEnd(&zstream);
					return null;
				}
				break;
			}
			else
				fileStream_.seek(dataLength, SeekPos.Current);

			// CRC
			fileStream_.seek(4, SeekPos.Current);
		}

		// ピクセルデータ一行のバイト数
		uint bytesPerPixelLine = width_ * bytePerPixel;

		// フィルタ解除後データの書き込み位置設定 ダミー行の直後から書き込みを開始する
		ubyte* writeCursor = &buffer[pitch];

		// フィルタ解除前データの読み込み位置設定 解凍データの先頭を設定
		ubyte* readCursor = &buffer[pitch + bytePerPixel * height_];

		// 1行ずつフィルタを解除する
		for (uint y; y < height_; ++y)
		{
			// フィルタタイプを取得
			ubyte filterType = *readCursor++;

			// 特に使用しないが一応フィルタタイプを書き込んでおく
			*writeCursor++ = filterType;

			// ダミーピクセル領域を0クリア
			for (uint i; i < bytePerPixel; ++i)
				*writeCursor++ = 0;

			// フィルタ処理を適用
			switch (filterType)
			{
			case 0:
				for (uint i; i < bytesPerPixelLine; ++i)
					writeCursor[i] = readCursor[i];
				break;

			case 1:
				for (uint i; i < bytesPerPixelLine; ++i)
				{
					ubyte b = readCursor[i];
					b += *(writeCursor + i - bytePerPixel);
					writeCursor[i] = b;
				}
				break;

			case 2:
				for (uint i; i < bytesPerPixelLine; ++i)
				{
					ubyte b = readCursor[i];
					b += *(writeCursor + i - pitch);
					writeCursor[i] = b;
				}
				break;

			case 3:
				for (uint i; i < bytesPerPixelLine; ++i)
				{
					ubyte b = readCursor[i];
					b += cast(ubyte)((*(writeCursor + i - bytePerPixel) + *(writeCursor + i - pitch)) / 2);
					writeCursor[i] = b;
				}
				break;

			case 4:
				for (uint i; i < bytesPerPixelLine; ++i)
				{
					ubyte a = *(writeCursor + i - bytePerPixel);
					ubyte b = *(writeCursor + i - pitch);
					ubyte c = *(writeCursor + i - pitch - bytePerPixel);
					int p = a + b - c;
					int pa = p - a;
					if (pa < 0)
						pa = -pa;
					int pb = p - b;
					if (pb < 0)
						pb = -pb;
					int pc = p - c;
					if (pc < 0)
						pc = -pc;

					if (pa <= pb && pa <= pc)
						c = a;
					else if (pb <= pc)
						c = b;

					b = readCursor[i];
					b += c;
					writeCursor[i] = b;
				}
				break;

			// 未定義のフィルタなのでエラーにしてもいいが、一応0と同じ処理をしておく
			default:
				for (uint i; i < bytesPerPixelLine; ++i)
					writeCursor[i] = readCursor[i];
				break;
			}

			// カーソルを次の行へ移動
			writeCursor += bytesPerPixelLine;
			readCursor += bytesPerPixelLine;
		}

		return buffer;
	}
}
