module jp.ku6.windows.game.application;

import core.thread;
import std.path;
import std.windows.charset;
import jp.ku6.windows.win32api.windef;
import jp.ku6.windows.win32api.winbase;
import jp.ku6.windows.win32api.winuser;
import jp.ku6.windows.win32api.objbase;
import jp.ku6.windows.win32api.mmsystem;

class GameError : Error
{
	this(string message)
	{
		super(message);
	}
}

/++
 + ウインドウメッセージ受信インターフェース
 +
 + ウインドウメッセージが必要なクラスはこのインターフェースを継承する
 +/
interface WindowMessageListener
{
	/++
	 + ウインドメッセージの処理
	 +
	 + Params:
	 +  windowHandle = ウインドウハンドル
	 +  message = ウインドウメッセージ
	 +  wParam = メッセージのパラメータ
	 +  lParam = メッセージのパラメタ
	 +/
	bool procedure(HWND windowHandle, UINT message, WPARAM wParam, LPARAM lParam);
}

/++
 + メインウインドウ
 +
 + アプリケーションに対して必ず一つ作成するメインウインドウ
 +/
class Window : WindowMessageListener
{
	@property
	{
		/// ウインドウハンドルを取得
		const
		HWND handle()
		{
			return windowHandle_;
		}

		/// アクティブ状態フラグ
		const
		bool active()
		{
			return active_;
		}

		/// 最小化状態フラグ
		const
		bool minimum()
		{
			return minimum_;
		}

		/// 左上点X座標
		const
		int x()
		{
			return x_;
		}

		void x(int i)
		{
			x_ = i;
			positionCenter_ = false;
			changePosition_ = true;
		}

		/// 左上点Y座標
		const
		int y()
		{
			return y_;
		}

		void y(int i)
		{
			y_ = i;
			positionCenter_ = false;
			changePosition_ = true;
		}

		/// 幅
		const
		int width()
		{
			return width_;
		}

		/// 高さ
		const
		int height()
		{
			return height_;
		}

		/// クライアント領域幅
		const
		int clientWidth()
		{
			return clientWidth_;
		}

		void clientWidth(int width)
		{
			clientWidth_ = width;
			changePosition_ = true;
		}

		/// クライアント領域高さ
		const
		int clientHeight()
		{
			return clientHeight_;
		}

		void clientHeight(int height)
		{
			clientHeight_ = height;
			changePosition_ = true;
		}

		/// フレーム有り無しフラグ
		const
		bool frame()
		{
			return frame_;
		}

		void frame(bool f)
		{
			if (frame_ != f)
			{
				frame_ = f;
				changeStyle_ = true;
				changePosition_ = true;
			}
		}
	}

	/++
	 + ウインドウをデスクトップ中央に配置する
	 +/
	void centering()
	{
		positionCenter_ = true;
		changePosition_ = true;
	}

	/++
	 + ウインドウの状態を画面に反映する
	 +
	 + ウインドウの位置、サイズ、表示項目等を変更した後、
	 + この関数を呼び出すことで実際に反映される
	 +/
	void refresh()
	{
		// ウインドウをいったん非表示に
		ShowWindow(windowHandle_, SW_HIDE);

		// フレームは有る？
		if (frame_)
		{
			// フレームとキャプションの分を加味する。本当はメニューとかもやるべき
			int clientWidth = clientWidth_ + GetSystemMetrics(SM_CXDLGFRAME) * 2;
			int clientHeight = clientHeight_ + GetSystemMetrics(SM_CYDLGFRAME) * 2 + GetSystemMetrics(SM_CYCAPTION);

			// センタリング判定
			if (positionCenter_)
			{
				x_ = (GetSystemMetrics(SM_CXSCREEN) - clientWidth) >> 1;
				y_ = (GetSystemMetrics(SM_CYSCREEN) - clientHeight) >> 1;
			}

			// スタイルを変更
			if (changeStyle_)
				SetWindowLong(windowHandle_, GWL_STYLE, WS_CAPTION | WS_SYSMENU | WS_MINIMIZEBOX);

			// サイズ、位置を変更
			if (changePosition_)
				SetWindowPos(windowHandle_, HWND_NOTOPMOST, x_, y_, clientWidth, clientHeight, 0);
		}
		else
		{
			// センタリング判定
			if (positionCenter_)
			{
				x_ = (GetSystemMetrics(SM_CXSCREEN) - clientWidth_) >> 1;
				y_ = (GetSystemMetrics(SM_CYSCREEN) - clientHeight_) >> 1;
			}

			// スタイルを変更
			if (changeStyle_)
				SetWindowLong(windowHandle_, GWL_STYLE, WS_POPUP);

			// サイズ、位置を変更
			if (changePosition_)
				SetWindowPos(windowHandle_, HWND_TOPMOST, x_, y_, clientWidth_, clientHeight_, 0);
		}

		// ウインドウを表示
		ShowWindow(windowHandle_, SW_SHOWNA);

		// フラグを初期化
		positionCenter_ = false;
		changeStyle_ = false;
		changePosition_ = false;
	}

	bool procedure(HWND windowHandle, UINT message, WPARAM wParam, LPARAM lParam)
	{
		switch (message) 
		{
		// ウインドウのアクティブ状態の変更時の処理
		case WM_ACTIVATE:
			active_ = LOWORD(wParam) != WA_INACTIVE;
			break;

		// ウインドウのサイズが変更されたときの処理
		case WM_SIZE:
			if (wParam == SIZE_MINIMIZED)
			{
				active_ = false;
				minimum_ = true;
			}
			else if (wParam == SIZE_RESTORED)
			{
				active_ = true;
				minimum_ = false;
			}
			break;

		default:
			break;
		}

		return true;
	}

private:

	// アクティブ状態フラグ
	// ウインドウにフォーカスが当たった状態の間trueになる
	bool active_ = true;

	// 最小化状態フラグ
	// ウインドウが最小化されている間trueになる
	bool minimum_ = false;

	// センタリングフラグ
	// このフラグがtrueの時にrefreshするとセンタリングされる
	bool positionCenter_ = false;

	// 形式変更フラグ
	// ウインドウのサイズや、フレームの有無を変更されると、refreshされるまでtrueになる
	bool changeStyle_ = false;

	// 位置変更フラグ
	// ウインドウの位置を変更されると、refreshされるまでtrueになる
	bool changePosition_ = false;

	// フレーム有り無しフラグ
	bool frame_ = true;

	// ウインドウハンドル
	HWND windowHandle_;

	// ウインドウ左上点のデスクトップ上X座標
	int x_;

	// ウインドウ左上点のデスクトップ上Y座標
	int y_;

	// ウインドウの幅
	int width_;

	// ウインドウの高さ
	int height_;

	// ウインドウクライアント領域の幅
	int clientWidth_;

	// ウインドウクライアント領域の高さ
	int clientHeight_;

	// ウインドウクラス名
	string windowClassName_;

	// ウインドウキャプション文字列
	string caption_;

	// メニューリソース名
	string menuResourceName_;

	// アイコンリソース名
	string iconResourceName_;

	// カーソルリソース名
	string cursorResourceName_;

	/+
	 + コンストラクタ
	 +
	 + Params:
	 +  instanceHandle = アプリケーション起動時に渡されるインスタンスハンドル
	 +  procedure = ウインドウメッセージを受け取るコールバック関数のポインタ
	 +  clientWidht = クライアント領域幅
	 +  clientHeight = クライアント領域高さ
	 +  windowClassName = ウインドウクラス名
	 +  caption = ウインドウのキャプションに表示する文字列
	 +  manuResourceName = メニュー用リソース名
	 +  iconResourceName = アイコンリソース名
	 +  cursorResourceName = カーソルリソース名
	 +/
	this(HINSTANCE instanceHandle, WNDPROC procedure, int clientWidth, int clientHeight, string windowClassName, string caption, string menuResourceName, string iconResourceName, string cursorResourceName)
	{
		// 引数で指定された情報を保存
		clientWidth_ = clientWidth;
		clientHeight_ = clientHeight;
		windowClassName_ = windowClassName is null ? null : windowClassName ~ '\0';
		caption_ = caption is null ? null : caption ~ '\0';
		menuResourceName_ = menuResourceName is null ? null : menuResourceName ~ '\0';
		iconResourceName_ = iconResourceName is null ? null : iconResourceName ~ '\0';
		cursorResourceName_ = cursorResourceName is null ? null : cursorResourceName ~ '\0';

		// ウィンドウクラスの登録
		WNDCLASSEX wndclass;
		wndclass.cbSize = WNDCLASSEX.sizeof;
		wndclass.style = 0;
		wndclass.lpfnWndProc = procedure;
		wndclass.cbClsExtra = 0;
		wndclass.cbWndExtra = 0;
		wndclass.hInstance = instanceHandle;
		wndclass.hbrBackground = cast(HBRUSH)(COLOR_WINDOW + 1);
		wndclass.lpszClassName = windowClassName_.ptr;

		// メニュー
		wndclass.lpszMenuName = menuResourceName_.ptr;

		// アイコン
		if (iconResourceName_ is null)
		{
			wndclass.hIcon = LoadIcon(null, null);
			wndclass.hIconSm = LoadIcon(null, null);
		}
		else
		{
			wndclass.hIcon = LoadIcon(instanceHandle, iconResourceName_.ptr);
			wndclass.hIconSm = LoadIcon(instanceHandle, iconResourceName_.ptr);
		}

		// カーソル
		if (cursorResourceName_ is null)
			wndclass.hCursor = LoadCursor(null, IDC_ARROW);
		else
			wndclass.hCursor = LoadCursor(instanceHandle, cursorResourceName_.ptr);

		// ウインドウクラスを登録
		RegisterClassEx(&wndclass);

		// 作成時はウインドウモードでよいと思われる
		width_ = clientWidth_ + GetSystemMetrics(SM_CXDLGFRAME) * 2;
		height_ = clientHeight_ + GetSystemMetrics(SM_CYDLGFRAME) * 2 + GetSystemMetrics(SM_CYCAPTION);
		if (wndclass.lpszMenuName !is null)
			height_ += GetSystemMetrics(SM_CYMENU);

		x_ = (GetSystemMetrics(SM_CXSCREEN) - width_) >> 1;
		y_ = (GetSystemMetrics(SM_CYSCREEN) - height_) >> 1;

		windowHandle_ = CreateWindowEx(0, wndclass.lpszClassName, caption_.ptr, WS_OVERLAPPED | WS_SYSMENU | WS_MINIMIZEBOX, x_, y_, width_, height_, null, null, instanceHandle, null);

		ShowWindow(windowHandle_, SW_SHOW);
		UpdateWindow(windowHandle_);
	}

	/+
	 + デストラクタ
	 +/
	~this()
	{
		// ウインドウ破棄
		DestroyWindow(windowHandle_);
	}
}

/++
 + アプリケーション本体クラス
 +
 + アプリケーション内でRootとなるクラス
 +/
class Application : WindowMessageListener
{
	/++
	 + 初期化処理
	 +
	 + アプリケーション開始時に1度だけ呼び出す
	 + shutdownと対にして使う
	 +
	 + Params:
	 +  instanceHandle = アプリケーション起動時に渡されるインスタンスハンドル
	 +  clientWidht = クライアント領域幅
	 +  clientHeight = クライアント領域高さ
	 +  windowClassName = ウインドウクラス名
	 +  caption = ウインドウのキャプションに表示する文字列
	 +  manuResourceName = メニュー用リソース名
	 +  iconResourceName = アイコンリソース名
	 +  cursorResourceName = カーソルリソース名
	 +/
	static
	void startup(HINSTANCE instanceHandle, int clientWidth, int clientHeight, string windowClassName, string caption, string menuResourceName, string iconResourceName, string cursorResourceName)
	{
		// ウインドウの作成
		Window window = new Window(instanceHandle, &callBackProcedure, clientWidth, clientHeight, windowClassName, caption, menuResourceName, iconResourceName, cursorResourceName);

		// インスタンスの生成
		instance_ = new Application(instanceHandle, window);
	}

	/++
	 + 終了処理
	 +/
	static
	void shutdown()
	{
		delete instance_;
	}

	@property
	{
		/// インスタンス
		static
		Application instance()
		{
			return instance_;
		}

		/// インスタンスハンドル
		const
		HINSTANCE instanceHandle()
		{
			return instanceHandle_;
		}

		/// ウインドウ
		Window window()
		{
			return window_;
		}
	}

	/++
	 + アプリケーションの2重起動チェック
	 +
	 + Params:
	 +  mutexName = 同一判定用文字列
	 +
	 + Returns:
	 +  trueを返した場合、既に同一アプリケーションが起動している
	 +/
	bool isStarted(string mutexName)
	{
		mutexName_ = mutexName ~ "\0";
		mutexHandle_ = CreateMutex(null, 0, mutexName_.ptr);
		return WaitForSingleObject(mutexHandle_, 0) == WAIT_OBJECT_0;
	}

	/++
	 + アプリケーションの生存確認
	 +/
	bool isValid()
	{
		MSG message;

		// 1フレームに一回メッセージを処理する
		if (PeekMessage(&message, null, 0, 0, PM_REMOVE) != 0)
		{
			TranslateMessage(&message);
			DispatchMessage(&message);
		}

		if (window_.minimum)
		{
			// ウインドウが最小化されているときはメッセージの処理のみを回す
			while (valid_ && window_.minimum)
			{
				if (PeekMessage(&message, null, 0, 0, PM_REMOVE) != 0)
				{
					TranslateMessage(&message);
					DispatchMessage(&message);
				}
				else
					WaitMessage();
			}

			lastTime_ = timeGetTime();
		}
		else
		{	
			// Wait処理を入れる

			uint now = timeGetTime();

			if (now - lastTime_ < oneFrameTime_)
			{
				core.thread.Thread.sleep((oneFrameTime_ - (now - lastTime_)) * 10000);
				lastTime_ += oneFrameTime_;
			}
			else
			{
				// とりあえず超過時間はカットする形で
				lastTime_ = now;
			}
		}

		return valid_;
	}

	/++
	 + アプリケーションを終了
	 +
	 + この関数の呼び出し以降、isValidがfalseを返すようになるので、
	 + それを持ってループを抜けるようにする
	 +/
	void exit()
	{
		valid_ = false;
	}

	/++
	 + ウインドウメッセージ配信先を登録
	 +
	 + Params:
	 +  listener = ウインドウメッセージ配信先
	 +/
	void addWindowMessageListener(WindowMessageListener listener)
	{
		listeners_ ~= listener;
	}

	/++
	 + ウインドウメッセージ配信先を削除
	 +
	 + params:
	 +  listener = ウインドウメッセージ配信先
	 +/
	void removeWindowMessageListener(WindowMessageListener listener)
	{
		for (uint i; i < listeners_.length - 1; ++i)
		{
			if (listeners_[i] is listener)
			{
				for (; i < listeners_.length - 1; ++i)
					listeners_[i] = listeners_[i + 1];
			}
		}

		listeners_.length = listeners_.length - 1;
	}

	bool procedure(HWND windowHandle, UINT message, WPARAM wParam, LPARAM lParam)
	{
		if (message == WM_CLOSE)
		{
			exit();
			return false;
		}
		else
			return true;
	}

private:

	// singleton
	static
	Application instance_;

	// 生存フラグ
	bool valid_ = true;

	// 1フレームに使用する時間
	uint oneFrameTime_ = 17;

	// 前回取得時の時間
	uint lastTime_;
	
	// mutex名
	string mutexName_;

	// mutexハンドル
	HANDLE mutexHandle_;

	// インスタンスハンドル
	HINSTANCE instanceHandle_;

	// メインウインドウ
	Window window_;

	// ウインドウメッセージ配信先
	WindowMessageListener[] listeners_;

	// ウインドウメッセージ配信でコールバックされる関数
	extern (Windows)
	static
	LRESULT callBackProcedure(HWND windowHandle, UINT message, WPARAM wParam, LPARAM lParam)
	{
		if (instance_ !is null)
		{
			foreach (listener; instance_.listeners_)
			{
				if (!listener.procedure(windowHandle, message, wParam, lParam))
					return 0;
			}
		}

		return DefWindowProc(windowHandle, message, wParam, lParam);
	}

	/+
	 + コンストラクタ
	 +
	 + instanceHandle = インスタンスハンドル
	 + window = メインウインドウ
	 +/
	this(HINSTANCE instanceHandle, Window window)
	{
		instanceHandle_ = instanceHandle;
		window_ = window;

		CoInitialize(null);

		timeBeginPeriod(1);
		lastTime_ = timeGetTime();

		addWindowMessageListener(this);
		addWindowMessageListener(window_);
	}

	/+
	 + デストラクタ
	 +/
	~this()
	{
		removeWindowMessageListener(window_);
		removeWindowMessageListener(this);

		if (mutexHandle_ != null)
		{
			ReleaseMutex(mutexHandle_);
			CloseHandle(mutexHandle_);
		}

		timeEndPeriod(1);

		CoUninitialize();

		delete window_;

		PostQuitMessage(0);
	}
}

/++
 + 実行ファイルが置かれたディレクトリを取得
 +
 + Returns:
 +  実行ファイルが置かれたディレクトリのパス
 +/
string executeDirectory(HINSTANCE hInstance)
{
	char[MAX_PATH] selfFilePath;
	GetModuleFileName(hInstance, selfFilePath.ptr, selfFilePath.length);

	uint i;
	for (; selfFilePath[i] != 0; ++i) {}

	return dirname(selfFilePath[0 .. i]).idup;
}

/++
 + Windowsアプリケーション コマンドライン引数を解析する
 +
 + Params:
 +  commandLine = WinMainの引数
 +
 + Returns:
 +  引数の配列
 +/
string[] parseCommandLine(LPSTR commandLine)
{
	string[] result = [];

	if (commandLine is null || commandLine[0] == 0)
		return result;

	for (uint cursor;;)
	{
		// 区切り文字を設定
		char separator;
		if (commandLine[cursor] == '"')
		{
			separator = '"';

			++cursor;
			if (commandLine[cursor] == 0)
				break;
		}
		else
			separator = ' ';

		// 文字列長を取得
		uint end = cursor;
		for (; commandLine[end] != separator && commandLine[end] != 0; ++end) {}
		uint stringLength = end - cursor;

		// バッファを確保
		char[] buffer = new char[stringLength + 1];

		// 文字列をコピー
		for (uint i; i < stringLength; ++i)
			buffer[i] = commandLine[cursor++];
		buffer[stringLength] = 0;

		// sjis -> utf8変換
		result ~= std.windows.charset.fromMBSz(cast(immutable(char)*)buffer); 

		// 文字列終端を確認
		if (commandLine[cursor] == 0)
			break;

		// "でくくってあった場合、"の次にカーソルを移動
		if (separator == '"')
		{
			++cursor;
			if (commandLine[cursor] == 0)
				break;
		}

		// 空白をスキップ
		for (; commandLine[cursor] == ' ' && commandLine[cursor] != 0; ++cursor) {}
		if (commandLine[cursor] == 0)
			break;
	}

	return result;
}

