module jp.ku6.windows.game.input;

import jp.ku6.windows.win32api.windef;
import jp.ku6.windows.win32api.winbase;
import jp.ku6.windows.win32api.winuser;
import jp.ku6.windows.win32api.objbase;
import jp.ku6.windows.win32api.wtypes;
import jp.ku6.windows.directx.dinput;
import jp.ku6.windows.game.application;

/++
 + 入力機器制御クラス
 +
 + GraphicsやSoundと違って、入力機器は使える上限が決まっている
 + (キーボード1個、マウス1個、ジョイパッド刺してある個数)
 + よって、それらのデバイスはnewで作成するのではなく、
 + Inputのinstanceから取得する形とする
 +/
class Input
{
	/++
	 + 初期化処理
	 +/
	static
	void startup()
	{
		instance_ = new Input;
	}

	/++
	 + 終了処理
	 +/
	static
	void shutdown()
	{
		delete instance_;
	}

	@property
	{
		/// singleton
		static
		Input instance()
		{
			return instance_;
		}

		/++
		 + キーボードオブジェクト
		 +/
		const(Keyboard) keyboard()
		{
			if (!keyboardInitialize_)
			{
				keyboardInitialize_ = true;
				keyboard_ = new Keyboard;
			}
			return keyboard_;
		}

		/++
		 + マウスオブジェクト
		 +/
		const(Mouse) mouse()
		{
			if (!mouseInitialize_)
			{
				mouseInitialize_ = true;
				mouse_ = new Mouse;
			}
			return mouse_;
		}

		/++
		 + ジョイパッドオブジェクト(複数)
		 +/
		const(Joypad[]) joypads()
		{
			if (!joypadInitialize_)
			{
				joypadInitialize_ = true;
				if (FAILED(Input.instance.directInput_.EnumDevices(DI8DEVCLASS_GAMECTRL, &enumerateDeviceCallback, cast(void*)&joypads_, DIEDFL_ATTACHEDONLY)))
					throw new GameError("Failed create Direct input device");
			}
			return joypads_;
		}
	}

	/++
	 + 機器の状態を更新する
	 +/
	void refresh()
	{
		if (keyboard_ !is null)
			keyboard_.refresh();

		if (mouse_ !is null)
			mouse_.refresh();

		foreach (joypad; joypads_)
			joypad.refresh();
	}

private:

	// singleton
	static
	Input instance_;

	// Direct input
	IDirectInput8 directInput_;

	// キーボード初期化フラグ(失敗時にも初期化試行したことがわかるように)
	bool keyboardInitialize_ = false;

	// マウス初期化フラグ
	bool mouseInitialize_ = false;

	// ジョイパッド初期化フラグ
	bool joypadInitialize_ = false;

	// キーボード
	Keyboard keyboard_;

	// マウス
	Mouse mouse_;

	// ジョイパッド
	Joypad[] joypads_;

	/+
	 + デバイス列挙時のコールバック処理
	 +/
	extern (Windows)
	static
	BOOL enumerateDeviceCallback(const(DIDEVICEINSTANCE)* device, LPVOID context)
	{
		auto joypad = new Joypad(device);
		if (joypad.directInputDevice_ !is null)
			*(cast(Joypad[]*)(context)) ~= joypad;

		return DIENUM_CONTINUE;
	}

	/+
	 + コンストラクタ
	 +/
	this()
	{
		if (FAILED(CoCreateInstance(&CLSID_DirectInput8, null, CLSCTX.CLSCTX_INPROC_SERVER, &IID_IDirectInput8, cast(VOID**)&directInput_)))
			throw new GameError("Failed get DirectInput");
		scope(failure)
		{
			directInput_.Release();
			directInput_ = null;
		}

		if (FAILED(directInput_.Initialize(Application.instance.instanceHandle, DIRECTINPUT_VERSION)))
			throw new GameError("Failed initialize Direct input");
	}

	/+
	 + デストラクタ
	 +/
	~this()
	{
		if (keyboard_ !is null)
			delete keyboard_;

		if (mouse_ !is null)
			delete mouse_;

		foreach (joypad; joypads_)
			delete joypad;

		// Direct inputの解放
		if (directInput_ !is null)
			directInput_.Release();
	}
}

/++
 + キーボードクラス
 +/
class Keyboard
{
	/++
	 + キーボードのキーの定数
	 +/
	enum Key
	{
		ESCAPE       = 0x01,
		N1           = 0x02,
		N2           = 0x03,
		N3           = 0x04,
		N4           = 0x05,
		N5           = 0x06,
		N6           = 0x07,
		N7           = 0x08,
		N8           = 0x09,
		N9           = 0x0A,
		N0           = 0x0B,
		MINUS        = 0x0C,
		EQUALS       = 0x0D,
		BACK         = 0x0E,
		TAB          = 0x0F,
		Q            = 0x10,
		W            = 0x11,
		E            = 0x12,
		R            = 0x13,
		T            = 0x14,
		Y            = 0x15,
		U            = 0x16,
		I            = 0x17,
		O            = 0x18,
		P            = 0x19,
		LBRACKET     = 0x1A,
		RBRACKET     = 0x1B,
		RETURN       = 0x1C,
		LCONTROL     = 0x1D,
		A            = 0x1E,
		S            = 0x1F,
		D            = 0x20,
		F            = 0x21,
		G            = 0x22,
		H            = 0x23,
		J            = 0x24,
		K            = 0x25,
		L            = 0x26,
		SEMICOLON    = 0x27,
		APOSTROPHE   = 0x28,
		GRAVE        = 0x29,
		LSHIFT       = 0x2A,
		BACKSLASH    = 0x2B,
		Z            = 0x2C,
		X            = 0x2D,
		C            = 0x2E,
		V            = 0x2F,
		B            = 0x30,
		N            = 0x31,
		M            = 0x32,
		COMMA        = 0x33,
		PERIOD       = 0x34,
		SLASH        = 0x35,
		RSHIFT       = 0x36,
		MULTIPLY     = 0x37,
		LMENU        = 0x38,
		SPACE        = 0x39,
		CAPITAL      = 0x3A,
		F1           = 0x3B,
		F2           = 0x3C,
		F3           = 0x3D,
		F4           = 0x3E,
		F5           = 0x3F,
		F6           = 0x40,
		F7           = 0x41,
		F8           = 0x42,
		F9           = 0x43,
		F10          = 0x44,
		NUMLOCK      = 0x45,
		SCROLL       = 0x46,
		NUMPAD7      = 0x47,
		NUMPAD8      = 0x48,
		NUMPAD9      = 0x49,
		SUBTRACT     = 0x4A,
		NUMPAD4      = 0x4B,
		NUMPAD5      = 0x4C,
		NUMPAD6      = 0x4D,
		ADD          = 0x4E,
		NUMPAD1      = 0x4F,
		NUMPAD2      = 0x50,
		NUMPAD3      = 0x51,
		NUMPAD0      = 0x52,
		DECIMAL      = 0x53,
		OEM_102      = 0x56,
		F11          = 0x57,
		F12          = 0x58,
		F13          = 0x64,
		F14          = 0x65,
		F15          = 0x66,
		KANA         = 0x70,
		ABNT_C1      = 0x73,
		CONVERT      = 0x79,
		NOCONVERT    = 0x7B,
		YEN          = 0x7D,
		ABNT_C2      = 0x7E,
		NUMPADEQUALS = 0x8D,
		PREVTRACK    = 0x90,
		AT           = 0x91,
		COLON        = 0x92,
		UNDERLINE    = 0x93,
		KANJI        = 0x94,
		STOP         = 0x95,
		AX           = 0x96,
		UNLABELED    = 0x97,
		NEXTTRACK    = 0x99,
		NUMPADENTER  = 0x9C,
		RCONTROL     = 0x9D,
		MUTE         = 0xA0,
		CALCULATOR   = 0xA1,
		PLAYPAUSE    = 0xA2,
		MEDIASTOP    = 0xA4,
		VOLUMEDOWN   = 0xAE,
		VOLUMEUP     = 0xB0,
		WEBHOME      = 0xB2,
		NUMPADCOMMA  = 0xB3,
		DIVIDE       = 0xB5,
		SYSRQ        = 0xB7,
		RMENU        = 0xB8,
		PAUSE        = 0xC5,
		HOME         = 0xC7,
		UP           = 0xC8,
		PRIOR        = 0xC9,
		LEFT         = 0xCB,
		RIGHT        = 0xCD,
		END          = 0xCF,
		DOWN         = 0xD0,
		NEXT         = 0xD1,
		INSERT       = 0xD2,
		DELETE       = 0xD3,
		LWIN         = 0xDB,
		RWIN         = 0xDC,
		APPS         = 0xDD,
		POWER        = 0xDE,
		SLEEP        = 0xDF,
		WAKE         = 0xE3,
		WEBSEARCH    = 0xE5,
		WEBFAVORITES = 0xE6,
		WEBREFRESH   = 0xE7,
		WEBSTOP      = 0xE8,
		WEBFORWARD   = 0xE9,
		WEBBACK      = 0xEA,
		MYCOMPUTER   = 0xEB,
		MAIL         = 0xEC,
		MEDIASELECT  = 0xED,
		BACKSPACE    = BACK,
		NUMPADSTAR   = MULTIPLY,
		LALT         = LMENU,
		CAPSLOCK     = CAPITAL,
		NUMPADMINUS  = SUBTRACT,
		NUMPADPLUS   = ADD,
		NUMPADPERIOD = DECIMAL,
		NUMPADSLASH  = DIVIDE,
		RALT         = RMENU,
		UPARROW      = UP,
		PGUP         = PRIOR,
		LEFTARROW    = LEFT,
		RIGHTARROW   = RIGHT,
		DOWNARROW    = DOWN,
		PGDN         = NEXT,
		CIRCUMFLEX   = PREVTRACK,
	}

	/++
	 + キーが押されているか判定
	 +
	 + 現在キーが押されているか判定する
	 +
	 + Params:
	 +  key = 判定するキー
	 +
	 + Returns:
	 +  true = 押されている
	 +  false = 押されていない
	 +/
	const
	bool isPress(Key key)
	{
		return (keyInformation_[bufferNumber_][key] & 0x80) != 0;
	}

	/++
	 + キーが押された瞬間か判定
	 +
	 + 1秒間に60フレーム等だと、普通に押しただけでisPressは複数フレームの間ONが返される
	 + この関数は押された最初のフレームだけONを返す
	 +
	 + Params:
	 +  key = 判定するキー
	 +
	 + Returns:
	 +  true = 押されている
	 +  false = 押されていない
	 +/
	const
	bool isPush(Key key)
	{
		return (keyInformation_[bufferNumber_][key] & 0x80) != 0 && (keyInformation_[bufferNumber_^1][key] & 0x80) == 0;
	}

private:

	// キー毎の入力状態バッファ
	immutable uint BUFFER_LENGTH = 256;

	// Direct input device
	IDirectInputDevice8 directInputDevice_;

	// 使用中のバッファ番号
	int bufferNumber_;

	// キー情報
	ubyte keyInformation_[2][BUFFER_LENGTH];

	/+
	 + コンストラクタ
	 +/
	this()
	{
		// Direct input deviceを作成
		if (FAILED(Input.instance.directInput_.CreateDevice(&GUID_SysKeyboard, &directInputDevice_, null))) 
			throw new GameError("Failed get Direct input device");
		scope(failure)
		{
			directInputDevice_.Release();
			directInputDevice_ = null;
		}

		// データ形式を設定
		if (FAILED(directInputDevice_.SetDataFormat(&c_dfDIKeyboard)))
			throw new GameError("Failed set data format");
		
		// モードを設定（フォアグラウンド＆非排他モード）
		if (FAILED(directInputDevice_.SetCooperativeLevel(Application.instance.window.handle, DISCL_NONEXCLUSIVE | DISCL_FOREGROUND)))
			throw new GameError("Failed set cooperative level");

		// 入力制御開始
		directInputDevice_.Acquire();
	}

	/+
	 + デストラクタ
	 +/
	~this()
	{
		if (directInputDevice_ !is null)
		{
			directInputDevice_.Unacquire(); 
			directInputDevice_.Release();
			directInputDevice_ = null;
		}
	}

	/+
	 + 機器状態の更新
	 +/
	void refresh()
	{
		if (directInputDevice_ is null)
			return;

		// バッファフリップ
		bufferNumber_ ^= 0x00000001;

		// データ取得
		if (FAILED(directInputDevice_.GetDeviceState(BUFFER_LENGTH, keyInformation_[bufferNumber_].ptr)))
		{
			directInputDevice_.Acquire();
			keyInformation_[bufferNumber_][0 .. $] = 0;
		}
	}
}

/++
 + マウス
 +/
class Mouse
{
	/++
	 + ボタンの定義
	 +/
	enum Button
	{
		LEFT,   /// 左ボタン
		RIGHT,  /// 右ボタン
		CENTER, /// 中ボタン
	}

	/++
	 + ボタンが押されているか判定
	 +
	 + 現在ボタンが押されているか判定する
	 +
	 + Params:
	 +  button = 判定するボタン
	 +
	 + Returns:
	 +  true = 押されている
	 +  false = 押されていない
	 +/
	const 
	bool isPress(Button button)
	{
		return (buttonPress_ & (1<<button)) != 0;
	}

	/++
	 + ボタンが押された瞬間か判定
	 +
	 + 1秒間に60フレーム等だと、普通に押しただけでisPressは複数フレームの間ONが返される
	 + この関数は押された最初のフレームだけONを返す
	 +
	 + Params:
	 +  button = 判定するボタン
	 +
	 + Returns:
	 +  true = 押されている
	 +  false = 押されていない
	 +/
	const
	bool isPush(Button button) 
	{
		return (buttonPush_ & (1<<button)) != 0;
	}

	/++
	 + X座標を取得
	 +
	 + Returns:
	 +  クライアント領域上のX座標
	 +/
	const
	int x()
	{
		return positionX_[0];
	}

	/++
	 + Y座標を取得
	 +
	 + Returns:
	 +  クライアント領域上のY座標
	 +/
	const
	int y()
	{
		return positionY_[0];
	}

	/++
	 + X軸移動量を取得
	 +
	 + Returns:
	 +  X軸移動量
	 +/
	const
	int moveX()
	{
		return positionX_[0] - positionX_[1];
	}

	/++
	 + Y軸移動量を取得
	 +
	 + Returns:
	 +  Y軸移動量
	 +/
	const
	int moveY()
	{
		return positionY_[0] - positionY_[1];
	}

	/++
	 + Z軸(ホイール)移動量を取得
	 +
	 + Returns:
	 +  Z軸(ホイール)移動量
	 +/
	const
	int moveZ()
	{
		return moveZ_;
	}

private:

	// ボタン数
	immutable uint BUTTON_COUNT = 3;

	// Direct Input Device
	IDirectInputDevice8 directInputDevice_;

	// X軸座標
	int[2] positionX_;

	// Y軸座業
	int[2] positionY_;

	// Z軸(ホイール)移動量
	int moveZ_;

	// ボタンが押されているかのフラグ
	uint buttonPress_;

	// ボタンが押された瞬間かのフラグ
	uint buttonPush_;

	/+
	 + コンストラクタ
	 +/
	this()
	{
		// Direct input deviceを作成
		if (FAILED(Input.instance.directInput_.CreateDevice(&GUID_SysMouse, &directInputDevice_, NULL)))
			throw new GameError("Failed get Direct input device");
		scope(failure)
		{
			directInputDevice_.Release();
			directInputDevice_ = null;
		}

		// データ形式を設定
		if (FAILED(directInputDevice_.SetDataFormat(&c_dfDIMouse2)))
			throw new GameError("Failed set data format");
		
		// モードを設定（フォアグラウンド＆非排他モード）
		if (FAILED(directInputDevice_.SetCooperativeLevel(Application.instance.window.handle, DISCL_NONEXCLUSIVE | DISCL_FOREGROUND)))
			throw new GameError("Failed set cooperative level");

		// 入力制御開始
		directInputDevice_.Acquire();
	}

	/+
	 + デストラクタ
	 +/
	~this()
	{
		if (directInputDevice_ !is null)
		{
			directInputDevice_.Unacquire(); 
			directInputDevice_.Release();
		}
	}

	/+
	 + 機器状態の更新
	 +/
	void refresh()
	{
		if (directInputDevice_ is null)
			return;

		// 情報をバックアップ
		uint beforeButtonPress = buttonPress_;
		positionX_[1] = positionX_[0];
		positionY_[1] = positionY_[0];

		// 現在のデータクリア
		buttonPress_ = 0;
		buttonPush_ = 0;

		// デバイスの直接データを取得する
		DIMOUSESTATE2 diMouseState;
		HRESULT hr = directInputDevice_.GetDeviceState(DIMOUSESTATE2.sizeof, &diMouseState);
		if (SUCCEEDED(hr))
		{
			POINT point;

			// 座標読み込み
			if (GetCursorPos(&point) != 0)
			{
				// クライアント領域左上からの相対値にする
				if (ScreenToClient(Application.instance.window.handle, &point) != 0)
				{
					positionX_[0] = point.x;
					positionY_[0] = point.y;
				}
			}

			// z軸移動量を取得
			moveZ_ = diMouseState.lZ;

			// 下からi bitめがi番のボタンのON,OFFフラグになる
			for (uint i; i < BUTTON_COUNT; ++i)
				buttonPress_ |= (((diMouseState.rgbButtons[i] >> 7) & 0x00000001) << i);

			buttonPush_ = buttonPress_ & (~beforeButtonPress);
		}
		else
		{
			directInputDevice_.Acquire();

			// 座標をクリア
			positionX_[0] = 0;
			positionY_[0] = 0;
			moveZ_ = 0;
		}
	}
}

/++
 + パッド
 +/
class Joypad
{
	/// ボタン
	enum Button
	{
		UP,
		DOWN,
		LEFT,
		RIGHT,
		A,
		B,
		C,
		D,
		E,
		F,
		G,
		H,
		I,
		J,
		K,
		L,
		LENGTH
	}

	/++
	 + ボタンが押されているか判定
	 +
	 + 現在ボタンが押されているか判定する
	 +
	 + Params:
	 +  key = 判定するボタン
	 +
	 + Returns:
	 +  true = 押されている
	 +  false = 押されていない
	 +/
	const
	bool isPress(Button button)
	{
		return (buttonPress & (1 << button)) != 0;
	}

	/++
	 + ボタンが押された瞬間か判定
	 +
	 + 1秒間に60フレーム等だと、普通に押しただけでisPressは複数フレームの間ONが返される
	 + この関数は押された最初のフレームだけONを返す
	 +
	 + Params:
	 +  button = 判定するボタン
	 +
	 + Returns:
	 +  true = 押されている
	 +  false = 押されていない
	 +/
	const
	bool isPush(Button button)
	{
		return (buttonPush & (1 << button)) != 0;
	}

private:

	// Direct Input device
	IDirectInputDevice8 directInputDevice_;

	// 現在ボタンが押されているかの情報
	uint buttonPress;

	// ボタンを押された瞬間の情報
	uint buttonPush;

	// 軸列挙時のコールバック処理
	extern (Windows)
	static
	BOOL enumerateAxesCallback(LPCDIDEVICEOBJECTINSTANCE device, LPVOID context)
	{
		DIPROPRANGE proprange;
		proprange.diph.dwSize = DIPROPRANGE.sizeof;
		proprange.diph.dwHeaderSize = DIPROPHEADER.sizeof;
		proprange.diph.dwObj = device.dwType;
		proprange.diph.dwHow = DIPH_BYID;
		proprange.lMin = -256;
		proprange.lMax = +256;

		IDirectInputDevice8 directInputDevice = cast(IDirectInputDevice8)context;
		if (FAILED(directInputDevice.SetProperty(DIPROP_RANGE, &proprange.diph)))
			return DIENUM_STOP;
		else
			return DIENUM_CONTINUE;
	}

	// コンストラクタ
	this(const(DIDEVICEINSTANCE)* device)
	{
		// ジョイスティック用にデバイスオブジェクトを作成
		if (SUCCEEDED(Input.instance.directInput_.CreateDevice(&device.guidInstance, &directInputDevice_, null)))
		{
			// データフォーマットを設定
			if (SUCCEEDED(directInputDevice_.SetDataFormat(&c_dfDIJoystick)))
			{
				// 協調モードを設定
				if (SUCCEEDED(directInputDevice_.SetCooperativeLevel(Application.instance.window.handle, DISCL_NONEXCLUSIVE | DISCL_FOREGROUND)))
				{
					// 軸を列挙して設定
					if (SUCCEEDED(directInputDevice_.EnumObjects(&enumerateAxesCallback, cast(LPVOID)directInputDevice_, DIDFT_AXIS)))
					{
						// Acquire
						if (SUCCEEDED(directInputDevice_.Acquire()))
						{
							return;
						}
					}
				}
			}
		}

		if (directInputDevice_ !is null)
			directInputDevice_.Release();
		directInputDevice_ = null;
	}

	// デストラクタ
	~this()
	{
		if (directInputDevice_ !is null)
		{
			directInputDevice_.Unacquire(); 
			directInputDevice_.Release();
			directInputDevice_ = null;
		}
	}

	// 機器状態の更新
	void refresh()
	{
		if (directInputDevice_ is null)
			return;

		// 前の入力情報をバックアップ
		uint oldButtonPress = buttonPress;

		// 情報をクリア
		buttonPress = 0;
		buttonPush = 0;

		// Poll失敗したら再Acquireする
		if (FAILED(directInputDevice_.Poll()))
		{
			if (FAILED(directInputDevice_.Acquire()))
				return;
		}

		// デバイスの直接データを取得する
		DIJOYSTATE joypadState;
		if (SUCCEEDED(directInputDevice_.GetDeviceState(DIJOYSTATE.sizeof, &joypadState)))
		{
			// X軸の入力をチェック
			if (joypadState.lX > 128)
				buttonPress |= 1 << Button.RIGHT;
			else if(joypadState.lX < -128)
				buttonPress |= 1 << Button.LEFT;

			// Y軸の傾きをチェック
			if (joypadState.lY > 128)
				buttonPress |= 1 << Button.DOWN;
			else if (joypadState.lY < -128)
				buttonPress |= 1 << Button.UP;

			// ボタンの入力をチェック
			for (int i = 4; i < Button.LENGTH; ++i)
				buttonPress |= ((joypadState.rgbButtons[i-4] >> 7) & 0x00000001) << i;

			// ボタンを押した瞬間かチェック
			buttonPush = buttonPress & (~oldButtonPress);
		}
	}
}

