module jp.ku6.pmdloader.DepthRenderTechnique;

import std.math;
import jp.ku6.windows.win32api.windows;
import jp.ku6.windows.directx.dxgiformat;
import jp.ku6.windows.directx.dxgitype;
import jp.ku6.windows.directx.dxgi;
import jp.ku6.windows.directx.d3dcommon;
import jp.ku6.windows.directx.d3d11;
import jp.ku6.windows.directx.d3dcompiler;
import jp.ku6.pmdloader.math;
import jp.ku6.pmdloader.pmd;

/++
 + 定数バッファ
 +/
struct Constants
{
	Matrix worldLightViewProjectionMatrix;
}

/++
 + 深度データをテクスチャに書き込むための処理
 +/
class DepthRenderTechnique
{
	/++
	 + コンストラクタ
	 +
	 + Params:
	 +  device = ID3D11Deviceインターフェース
	 +  D3DCompile = D3DCompile関数のポインタ
	 +  width = 深度バッファテクスチャの幅 大きいほど精度が上がる
	 +  height = 深度バッファテクスチャの高さ 大きいほど精度が上がる
	 +/
	this(ID3D11Device device, D3DCompileFunction D3DCompile, int width, int height)
	{
		// 深度データを書き込むテクスチャを生成
		D3D11_TEXTURE2D_DESC renderTextureDesc;
		renderTextureDesc.Width = width;
		renderTextureDesc.Height = height;
		renderTextureDesc.MipLevels = 1;
		renderTextureDesc.ArraySize = 1;
		renderTextureDesc.Format = DXGI_FORMAT.DXGI_FORMAT_R32_FLOAT;
		renderTextureDesc.SampleDesc.Count = 1;
		renderTextureDesc.SampleDesc.Quality = 0;
		renderTextureDesc.Usage = D3D11_USAGE.D3D11_USAGE_DEFAULT;
		renderTextureDesc.BindFlags = D3D11_BIND_FLAG.D3D11_BIND_SHADER_RESOURCE | D3D11_BIND_FLAG.D3D11_BIND_RENDER_TARGET;
		renderTextureDesc.CPUAccessFlags = 0;
		if (FAILED(device.CreateTexture2D(&renderTextureDesc, null, &renderTarget_)))
		{
			cleanup();
			throw new Error("Failed create depth stencil shadow texture");
		}

		// テクスチャをレンダリングターゲットとして扱う為のインターフェースを生成
		// (深度データを書き込む為)
		if (FAILED(device.CreateRenderTargetView(renderTarget_, null, &renderTargetView_)))
		{
			cleanup();
			throw new Error("Failed create depth stencil shadow texture view");
		}

		// テクスチャをシェーダリソースとして扱う為のインターフェースを生成
		// (書き込んだ深度データをピクセルシェーダから参照する為)
		if (FAILED(device.CreateShaderResourceView(renderTarget_, null, &renderTargetResourceView_)))
		{
			cleanup();
			throw new Error("Failed create depth stencil shadow texture resource view");
		}

		// 深度データのレンダリングに使う深度バッファ(ややこしい)
		// こいつをShader resourceとして使う方法があれば単純化出来るかも？
		D3D11_TEXTURE2D_DESC depthTextureDesc;
		depthTextureDesc.Width  = width;
		depthTextureDesc.Height = height;
		depthTextureDesc.MipLevels = 1;
		depthTextureDesc.ArraySize = 1;
		depthTextureDesc.Format = DXGI_FORMAT.DXGI_FORMAT_D32_FLOAT;
		depthTextureDesc.SampleDesc.Count = 1;
		depthTextureDesc.SampleDesc.Quality = 0;
		depthTextureDesc.Usage = D3D11_USAGE.D3D11_USAGE_DEFAULT;
		depthTextureDesc.BindFlags = D3D11_BIND_FLAG.D3D11_BIND_DEPTH_STENCIL;
		depthTextureDesc.CPUAccessFlags = 0;
		depthTextureDesc.MiscFlags = 0;

		ID3D11Texture2D depthBuffer;
		if (FAILED(device.CreateTexture2D(&depthTextureDesc, null, &depthBuffer)))
		{
			cleanup();
			throw new Error("Failed create depth buffer.");
		}

		// 深度ステンシルビューの作成
		D3D11_DEPTH_STENCIL_VIEW_DESC depthStencilViewDesc;
		depthStencilViewDesc.Format = depthTextureDesc.Format;
		depthStencilViewDesc.ViewDimension = D3D11_DSV_DIMENSION.D3D11_DSV_DIMENSION_TEXTURE2DMS;
		depthStencilViewDesc.Flags = 0;

		if (FAILED(device.CreateDepthStencilView(depthBuffer, &depthStencilViewDesc, &depthStencilView_)))
		{
			depthBuffer.Release();
			cleanup();
			throw new Error("Failed create depth stencil view.");
		}

		depthBuffer.Release();


		//--------------------------------------------------------------
		// シェーダコンパイルの準備

		// シェーダコンパイル後のデータとエラーを格納する
		ID3DBlob compiled;
		ID3DBlob errors;

		//--------------------------------------------------------------
		// Input Assembler ステージで使用するデータを準備

		// Vertex shader プログラムソースをコンパイル
		string vertexShaderSource = `
		cbuffer Constants : register(b0)
		{
			// ライトをカメラとしたときのワールドビュー射影変換行列
			float4x4 worldLightViewProjectionMatrix;
		};

		struct VS_INPUT
		{
			float4 pos : POSITION;
			float3 nor : NORMAL;  // norは使わないけど、通常のレンダリングと入力を合わせておく
			float2 uv : TEXCOORD;
		};

		struct PS_INPUT
		{
			float4 pos : SV_POSITION;
			float2 uv : TEXCOORD;
		};

		PS_INPUT VS_main(VS_INPUT input)
		{
			PS_INPUT output;
			output.pos = mul(input.pos, worldLightViewProjectionMatrix);
			output.uv = input.uv;
			return output;
		}
		`;
		if (FAILED(D3DCompile(vertexShaderSource.ptr, vertexShaderSource.length, null, null, null, "VS_main\0".ptr, "vs_4_0\0".ptr, 0, 0, &compiled, &errors)))
		{
			string message = "";
			if (errors !is null)
			{
				message = (cast(char*)errors.GetBufferPointer())[0 .. errors.GetBufferSize()].idup;
				errors.Release();
			}
			throw new Error("Failed compile vertex shader: " ~ message);
		}

		// Input assembler ステージの入力レイアウトを作成
		D3D11_INPUT_ELEMENT_DESC[3] vertexDescs;
		vertexDescs[0] = D3D11_INPUT_ELEMENT_DESC("POSITION\0".ptr, 0, DXGI_FORMAT.DXGI_FORMAT_R32G32B32_FLOAT, 0, 0, D3D11_INPUT_CLASSIFICATION.D3D11_INPUT_PER_VERTEX_DATA, 0);
		vertexDescs[1] = D3D11_INPUT_ELEMENT_DESC("NORMAL\0".ptr, 0, DXGI_FORMAT.DXGI_FORMAT_R32G32B32_FLOAT, 0, 12, D3D11_INPUT_CLASSIFICATION.D3D11_INPUT_PER_VERTEX_DATA, 0);
		vertexDescs[2] = D3D11_INPUT_ELEMENT_DESC("TEXCOORD\0".ptr, 0, DXGI_FORMAT.DXGI_FORMAT_R32G32_FLOAT, 0, 24, D3D11_INPUT_CLASSIFICATION.D3D11_INPUT_PER_VERTEX_DATA, 0);

		if (FAILED(device.CreateInputLayout(vertexDescs.ptr, vertexDescs.length, compiled.GetBufferPointer(), compiled.GetBufferSize(), &inputLayout_)))
		{
			compiled.Release();
			if (errors !is null)
				errors.Release();

			cleanup();
			throw new Error("Failed create input layout");
		}


		//--------------------------------------------------------------
		// Vertex Shader ステージで使用するデータを準備

		// Vertex shaderを作成
		if (FAILED(device.CreateVertexShader(compiled.GetBufferPointer(), compiled.GetBufferSize(), null, &vertexShader_)))
		{
			compiled.Release();
			if (errors !is null)
				errors.Release();
			cleanup();
			throw new Error("Failed create vertex shader");
		}

		compiled.Release();
		if (errors !is null)
			errors.Release();


		// 頂点シェーダ定数バッファを作成
		D3D11_BUFFER_DESC vertexShaderConstantsBufferDesc = D3D11_BUFFER_DESC(Constants.sizeof, D3D11_USAGE.D3D11_USAGE_DEFAULT, D3D11_BIND_FLAG.D3D11_BIND_CONSTANT_BUFFER, 0, 0);
		if (FAILED(device.CreateBuffer(&vertexShaderConstantsBufferDesc, null, &vertexShaderConstantsBuffer_)))
		{
			cleanup();
			throw new Error("Failed create constants buffer");
		}

		//--------------------------------------------------------------
		// Rasterizer ステージで使用するデータを準備

		// Viewportの設定
		viewPort_.Width = width;
		viewPort_.Height = height;
		viewPort_.MinDepth = 0.0f;
		viewPort_.MaxDepth = 1.0f;
		viewPort_.TopLeftX = 0;
		viewPort_.TopLeftY = 0;

		// Rasterizer state
		D3D11_RASTERIZER_DESC rasterizerDesc;
		rasterizerDesc.FillMode = D3D11_FILL_MODE.D3D11_FILL_SOLID;
		rasterizerDesc.CullMode = D3D11_CULL_MODE.D3D11_CULL_NONE;
		rasterizerDesc.DepthClipEnable = FALSE;
		rasterizerDesc.MultisampleEnable = FALSE;
		rasterizerDesc.DepthBiasClamp = 0;
		rasterizerDesc.SlopeScaledDepthBias = 0;
		if (FAILED(device.CreateRasterizerState(&rasterizerDesc, &rasterizerState_)))
		{
			cleanup();
			throw new Error("Failed create rasterizer state");
		}


		//--------------------------------------------------------------
		// Pixel Shader ステージで使用するデータを準備

		// テクスチャを使用しないPixel shaderを準備
		string polygonPixelShaderSource =
		`
			cbuffer Material : register(b0)
			{
				float3 diffuseColor;
				float alpha;
				float3 specularColor;
				float shininess;
				float3 ambientColor;
				float dummy;
			};

			struct PS_INPUT
			{
				float4 pos : SV_POSITION;
				float2 uv : TEXCOORD;
			};

			float4 PS_main(PS_INPUT input) : SV_Target
			{
				// モデルのアルファ値が0ならば表示しない
				if (alpha == 0)
					discard;

				// Z値出力する
				return input.pos.z;
			}
		`;
		if (FAILED(D3DCompile(polygonPixelShaderSource.ptr, polygonPixelShaderSource.length, null, null, null, "PS_main\0".ptr, "ps_4_0\0".ptr, 0, 0, &compiled, &errors)))
		{
			string message = "";
			if (errors !is null)
			{
				message = (cast(char*)errors.GetBufferPointer())[0 .. errors.GetBufferSize()].idup;
				errors.Release();
			}
			cleanup();
			throw new Error("Failed compile pixel shader: " ~ message);
		}

		if (FAILED(device.CreatePixelShader(compiled.GetBufferPointer(), compiled.GetBufferSize(), null, &polygonPixelShader_)))
		{
			compiled.Release();
			if (errors !is null)
				errors.Release();
			cleanup();
			throw new Error("Failed create pixel shader");
		}

		compiled.Release();
		if (errors !is null)
			errors.Release();


		// テクスチャを使用するPixel shaderを準備
		string texturePixelShaderSource =
		`
			cbuffer Material : register(b0)
			{
				float3 diffuseColor;
				float alpha;
				float3 specularColor;
				float shininess;
				float3 ambientColor;
				float dummy;
			};

			Texture2D tex : register(t0);

			SamplerState sam : register(s0);

			struct PS_INPUT
			{
				float4 pos : SV_POSITION;
				float2 uv : TEXCOORD;
			};

			float4 PS_main(PS_INPUT input) : SV_Target
			{
				// モデルのアルファ値が0ならば表示しない
				if (alpha == 0)
					discard;

				// テクスチャから取ってアルファ値が0ならば表示しない
				float4 pix = tex.Sample(sam, input.uv);
				if (pix.a == 0)
					discard;

				// Z値出力する
				return input.pos.z;
			}
		`;
		if (FAILED(D3DCompile(texturePixelShaderSource.ptr, texturePixelShaderSource.length, null, null, null, "PS_main\0".ptr, "ps_4_0\0".ptr, 0, 0, &compiled, &errors)))
		{
			string message = "";
			if (errors !is null)
			{
				message = (cast(char*)errors.GetBufferPointer())[0 .. errors.GetBufferSize()].idup;
				errors.Release();
			}
			cleanup();
			throw new Error("Failed compile pixel shader: " ~ message);
		}

		if (FAILED(device.CreatePixelShader(compiled.GetBufferPointer(), compiled.GetBufferSize(), null, &texturePixelShader_)))
		{
			compiled.Release();
			if (errors !is null)
				errors.Release();
			cleanup();
			throw new Error("Failed create pixel shader");
		}

		compiled.Release();
		if (errors !is null)
			errors.Release();

		// ピクセルシェーダ定数バッファを作成
		D3D11_BUFFER_DESC pixelShaderConstantsBufferDesc = D3D11_BUFFER_DESC(PmdFile.Material.sizeof, D3D11_USAGE.D3D11_USAGE_DEFAULT, D3D11_BIND_FLAG.D3D11_BIND_CONSTANT_BUFFER, 0, 0);
		if (FAILED(device.CreateBuffer(&pixelShaderConstantsBufferDesc, null, &pixelShaderConstantsBuffer_)))
		{
			cleanup();
			throw new Error("Failed create constants buffer");
		}

		D3D11_SAMPLER_DESC samplerDesc;
		samplerDesc.Filter = D3D11_FILTER.D3D11_FILTER_MIN_MAG_MIP_LINEAR;
		samplerDesc.AddressU = D3D11_TEXTURE_ADDRESS_MODE.D3D11_TEXTURE_ADDRESS_WRAP;
		samplerDesc.AddressV = D3D11_TEXTURE_ADDRESS_MODE.D3D11_TEXTURE_ADDRESS_WRAP;
		samplerDesc.AddressW = D3D11_TEXTURE_ADDRESS_MODE.D3D11_TEXTURE_ADDRESS_WRAP;
		samplerDesc.MipLODBias = 0.0f;
		samplerDesc.MaxAnisotropy = 1;
		samplerDesc.ComparisonFunc = D3D11_COMPARISON_FUNC.D3D11_COMPARISON_ALWAYS;
		samplerDesc.BorderColor[0] = 0.0f;
		samplerDesc.BorderColor[1] = 0.0f;
		samplerDesc.BorderColor[2] = 0.0f;
		samplerDesc.BorderColor[3] = 0.0f;
		samplerDesc.MinLOD = 0.0f;
		samplerDesc.MaxLOD = D3D11_FLOAT32_MAX;

		if (FAILED(device.CreateSamplerState(&samplerDesc, &samplerState_)))
		{
			cleanup();
		    throw new Error("Failed create sampler state");
		}


		//--------------------------------------------------------------
		// Output Merger ステージで使用するデータを準備

		// 深度ステンシルバッファの設定
		D3D11_DEPTH_STENCIL_DESC depthStencilDesc;
		depthStencilDesc.DepthEnable = TRUE;
		depthStencilDesc.DepthWriteMask = D3D11_DEPTH_WRITE_MASK.D3D11_DEPTH_WRITE_MASK_ALL;
		depthStencilDesc.DepthFunc = D3D11_COMPARISON_FUNC.D3D11_COMPARISON_LESS_EQUAL;
		depthStencilDesc.StencilEnable = FALSE;
		if (FAILED(device.CreateDepthStencilState(&depthStencilDesc, &depthStencilState_)))
		{
			cleanup();
			throw new Error("Failed create depth stencil state");
		}

		// アルファブレンドの設定
		D3D11_BLEND_DESC blendStateDesc;
		blendStateDesc.AlphaToCoverageEnable = FALSE;
		blendStateDesc.IndependentBlendEnable = FALSE;
		blendStateDesc.RenderTarget[0].BlendEnable = FALSE;
		blendStateDesc.RenderTarget[0].SrcBlend = D3D11_BLEND.D3D11_BLEND_ONE;
		blendStateDesc.RenderTarget[0].DestBlend = D3D11_BLEND.D3D11_BLEND_ZERO;
		blendStateDesc.RenderTarget[0].BlendOp = D3D11_BLEND_OP.D3D11_BLEND_OP_ADD;
		blendStateDesc.RenderTarget[0].SrcBlendAlpha = D3D11_BLEND.D3D11_BLEND_ONE;
		blendStateDesc.RenderTarget[0].DestBlendAlpha = D3D11_BLEND.D3D11_BLEND_ZERO;
		blendStateDesc.RenderTarget[0].BlendOpAlpha = D3D11_BLEND_OP.D3D11_BLEND_OP_ADD;
		blendStateDesc.RenderTarget[0].RenderTargetWriteMask = D3D11_COLOR_WRITE_ENABLE.D3D11_COLOR_WRITE_ENABLE_ALL;
		if (FAILED(device.CreateBlendState(&blendStateDesc, &blendState_)))
		{
			cleanup();
			throw new Error("Failed create blend state");
		}
	}

	/++
	 + デストラクタ
	 +
	 + アプリケーション終了時に呼び出される
	 +/
	~this()
	{
		cleanup();
	}

	/++
	 + 毎フレームの処理
	 +
	 + 一定時間毎に呼び出される
	 +
	 + Params:
	 +  deviceContext = ID3D11DeviceContextインターフェース
	 +  pmdFile = 描画するPMDデータ
	 +  worldLightViewProjectionMatrix = ワールドビュー射影変換行列
	 +
	 + Returns:
	 +  描画結果をシェーダリソースビューとして返す
	 +/
	ID3D11ShaderResourceView render(ID3D11DeviceContext deviceContext, PmdFile pmdFile, Matrix worldLightViewProjectionMatrix)
	{
		// ワールド変換行列が毎フレーム変動するので、変換行列は作った物を渡して貰う
		constants_.worldLightViewProjectionMatrix = worldLightViewProjectionMatrix;


		//--------------------------------------------------------------
		// パイプラインの各ステージの設定

		// Input Assember ステージに入力するデータ形式を設定
		deviceContext.IASetInputLayout(inputLayout_);

		// Input Assember ステージに入力されたデータの使用方法を設定
		deviceContext.IASetPrimitiveTopology(D3D11_PRIMITIVE_TOPOLOGY.D3D11_PRIMITIVE_TOPOLOGY_TRIANGLELIST);

		// Vertex Shader ステージに Vertex Shader を設定
		deviceContext.VSSetShader(vertexShader_, null, 0);

		// 定数バッファの内容を更新
		deviceContext.UpdateSubresource(vertexShaderConstantsBuffer_, 0, null, &constants_, 0, 0);
		
		// Vertex Shaderに定数バッファをセット
		deviceContext.VSSetConstantBuffers(0, 1, &vertexShaderConstantsBuffer_);

		// Rasterizer ステージにビューポートを設定
		deviceContext.RSSetViewports(1, &viewPort_);

		// Rasterizer ステージの設定
		deviceContext.RSSetState(rasterizerState_);

		// Output Merger ステージで最終的なレンダリングターゲットをバックバッファに設定
		deviceContext.OMSetRenderTargets(1, &renderTargetView_, depthStencilView_);

		// Output Merger ステージで深度ステンシルを設定
		deviceContext.OMSetDepthStencilState(depthStencilState_, 0);

		// Output Merger ステージでアルファブレンドを設定
		deviceContext.OMSetBlendState(blendState_, null, 0xffffffff);


		//--------------------------------------------------------------
		// 描画処理

		// 深度バッファをクリア
		deviceContext.ClearDepthStencilView(depthStencilView_, D3D11_CLEAR_FLAG.D3D11_CLEAR_DEPTH, 1.0f, 0);

		// レンダリングターゲットを一色で塗りつぶし
		float[4] clearColor = [1, 1, 1, 1]; // R32_FLOATだけど4成分必要っぽい
		deviceContext.ClearRenderTargetView(renderTargetView_, clearColor.ptr);

		// Input Assembler ステージに頂点をセット
		uint stride = PmdFile.Vertex.sizeof;
		uint offset = 0;
		deviceContext.IASetVertexBuffers(0, 1, &pmdFile.vertexBuffer, &stride, &offset);

		for (uint i; i < pmdFile.parts.length; ++i)
		{
			// Input Assembler ステージにインデックスをセット
			deviceContext.IASetIndexBuffer(pmdFile.parts[i].indexBuffer, DXGI_FORMAT.DXGI_FORMAT_R16_UINT, 0);

			// テクスチャ有無でシェーダを分ける
			if (pmdFile.parts[i].textureView is null)
			{
				// Pixel Shaderに Pixel Shader を設定
				deviceContext.PSSetShader(polygonPixelShader_, null, 0);
			}
			else
			{
				// Pixel Shaderに Pixel Shader を設定
				deviceContext.PSSetShader(texturePixelShader_, null, 0);

				// Pixel Shaderにサンプラをセット
				deviceContext.PSSetSamplers(0, 1, &samplerState_);

				// Pixel Shaderにテクスチャをセット
				deviceContext.PSSetShaderResources(0, 1, &pmdFile.parts[i].textureView);
			}

			// Pixel Shaderの定数バッファの内容を更新
			deviceContext.UpdateSubresource(pixelShaderConstantsBuffer_, 0, null, &pmdFile.parts[i].material, 0, 0);

			// Pixel Shaderに定数バッファをセット
			deviceContext.PSSetConstantBuffers(0, 1, &pixelShaderConstantsBuffer_);

			// パイプラインを実行して描画する
			deviceContext.DrawIndexed(pmdFile.parts[i].vertexCount, 0, 0);
		}

		return renderTargetResourceView_;
	}

private:

	// レンダリングターゲット
	ID3D11Texture2D renderTarget_;

	// Render target view
	ID3D11RenderTargetView renderTargetView_;

	// レンダリングターゲットをテクスチャとして使用する場合のインターフェース
	ID3D11ShaderResourceView renderTargetResourceView_;

	// 深度ステンシルビュー
	ID3D11DepthStencilView depthStencilView_;

	// Input layout
	ID3D11InputLayout inputLayout_;

	// Vertex shader
	ID3D11VertexShader vertexShader_;

	// 頂点シェーダ定数バッファ
	ID3D11Buffer vertexShaderConstantsBuffer_;

	// View port
	D3D11_VIEWPORT viewPort_;

	// Rasterizer設定
	ID3D11RasterizerState rasterizerState_;

	// テクスチャを使用しないPixel shader
	ID3D11PixelShader polygonPixelShader_;

	// テクスチャを使用するPixel shader
	ID3D11PixelShader texturePixelShader_;

	// ピクセルシェーダ定数バッファ
	ID3D11Buffer pixelShaderConstantsBuffer_;

	// サンプラ
	ID3D11SamplerState samplerState_;

	// 深度ステンシル設定
	ID3D11DepthStencilState depthStencilState_;

	// アルファブレンド設定
	ID3D11BlendState blendState_;

	// 定数バッファに送るデータ
	Constants constants_;

	/+
	 + 解放処理
	 +/
	void cleanup()
	{
		if (blendState_ !is null)
		{
			blendState_.Release();
			blendState_ = null;
		}

		if (depthStencilState_ !is null)
		{
			depthStencilState_.Release();
			depthStencilState_ = null;
		}

		if (samplerState_ !is null)
		{
			samplerState_.Release();
			samplerState_ = null;
		}

		if (pixelShaderConstantsBuffer_ !is null)
		{
			pixelShaderConstantsBuffer_.Release();
			pixelShaderConstantsBuffer_ = null;
		}

		if (texturePixelShader_ !is null)
		{
			texturePixelShader_.Release();
			texturePixelShader_ = null;
		}

		if (polygonPixelShader_ !is null)
		{
			polygonPixelShader_.Release();
			polygonPixelShader_ = null;
		}

		if (rasterizerState_ !is null)
		{
			rasterizerState_.Release();
			rasterizerState_ = null;
		}

		if (vertexShaderConstantsBuffer_ !is null)
		{
			vertexShaderConstantsBuffer_.Release();
			vertexShaderConstantsBuffer_ = null;
		}

		if (vertexShader_ !is null)
		{
			vertexShader_.Release();
			vertexShader_ = null;
		}

		if (inputLayout_ !is null)
		{
			inputLayout_.Release();
			inputLayout_ = null;
		}

		if (depthStencilView_ !is null)
		{
			depthStencilView_.Release();
			depthStencilView_ = null;
		}

		if (renderTargetResourceView_ !is null)
		{
			renderTargetResourceView_.Release();
			renderTargetResourceView_ = null;
		}

		if (renderTargetView_ !is null)
		{
			renderTargetView_.Release();
			renderTargetView_ = null;
		}

		if (renderTarget_ !is null)
		{
			renderTarget_.Release();
			renderTarget_ = null;
		}
	}
}

