module jp.ku6.pmdloader.pmd;

import std.algorithm;
import std.stream;
import std.path;
import jp.ku6.windows.win32api.windows;
import jp.ku6.windows.directx.dxgiformat;
import jp.ku6.windows.directx.d3d11;
import jp.ku6.bitmap.png;
import jp.ku6.bitmap.tga;
import jp.ku6.bitmap.bmp;

/++
 + PMDファイル操作クラス
 +/
class PmdFile
{
	/++
	 + 頂点情報
	 +/
	struct Vertex
	{
		/// 頂点座標
		float x;
		float y;
		float z;

		/// 法線ベクトル
		float nx;
		float ny;
		float nz;

		/// テクスチャ座標
		float u;
		float v;
	}

	/++
	 + 材質
	 +/
	struct Material
	{
		/// 拡散反射色
		float[3] diffuseColor;

		/// 透明度
		float alpha;

		/// 鏡面反射色
		float[3] specularColor;

		/// 鏡面反射度
		float shininess;

		/// 環境色
		float[3] ambientColor;

		float dummy;
	}

	/++
	 + モデルの各パーツ
	 +
	 + Materialはそのままシェーダに突っ込みたいので分離
	 + 一緒にするならシェーダに突っ込むデータを先頭に並べるとかすればよい
	 +/
	struct Part
	{
		/// 頂点数
		uint vertexCount;

		/// インデックスバッファ
		ID3D11Buffer indexBuffer;

		/// テクスチャビュー
		ID3D11ShaderResourceView textureView;

		/// トゥーンテクスチャの番号
		ubyte toonIndex;

		/// 
		ubyte edgeFlag;

		/// 材質
		Material material;
	}

	/// 頂点配列
	Vertex[] vertices;

	/// インデックス配列
	ushort[] indices;

	/// マテリアル配列
	Part[] parts;

	// 頂点バッファ
	ID3D11Buffer vertexBuffer;

	/++
	 + コンストラクタ
	 +
	 + Params:
	 +  filePath = 読み出すファイルのパス
	 +  ID3D11Device = Direct3D device
	 +/
	this(string filePath, ID3D11Device device)
	{
		// ファイルをオープン
		auto file = new BufferedFile(filePath);
		scope(exit) file.close();

		// PMDファイルの置いてあるディレクトリを取得
		string directory = std.path.dirname(filePath);

		// ヘッダ (ひとまず読み飛ばし)
		file.seek(3 + 4 + 20 + 256, SeekPos.Current);

		// 頂点数の読み出し
		uint vertexLength;
		file.read(vertexLength);

		// 頂点バッファを取得
		vertices = new Vertex[vertexLength];

		// 頂点バッファをロード
		for (uint i; i < vertexLength; ++i)
		{
			file.read(vertices[i].x);
			file.read(vertices[i].y);
			file.read(vertices[i].z);
			file.read(vertices[i].nx);
			file.read(vertices[i].ny);
			file.read(vertices[i].nz);
			file.read(vertices[i].u);
			file.read(vertices[i].v);

			file.seek(6, SeekPos.Current);
		}

		// 頂点バッファを作成
		D3D11_BUFFER_DESC vertexBufferDesc = D3D11_BUFFER_DESC(Vertex.sizeof * vertices.length, D3D11_USAGE.D3D11_USAGE_DEFAULT, D3D11_BIND_FLAG.D3D11_BIND_VERTEX_BUFFER, 0, 0);
		D3D11_SUBRESOURCE_DATA vertexSubsourceData = D3D11_SUBRESOURCE_DATA(vertices.ptr);
		if (FAILED(device.CreateBuffer(&vertexBufferDesc, &vertexSubsourceData, &vertexBuffer)))
		{
			cleanup();
			throw new Error("Failed create vertex buffer");
		}


		// インデックス数の読み出し
		uint indexLength;
		file.read(indexLength);

		// インデックスバッファを取得
		indices = new ushort[indexLength];

		// インデックスバッファをロード
		for (uint i; i < indexLength; ++i)
			file.read(indices[i]);


		// マテリアル数の読み出し
		uint partLength;
		file.read(partLength);

		// マテリアルを取得
		parts = new Part[partLength];

		// マテリアルをロード
		uint vertexIndex;
		for (uint i; i < partLength; ++i)
		{
			file.read(parts[i].material.diffuseColor[0]);
			file.read(parts[i].material.diffuseColor[1]);
			file.read(parts[i].material.diffuseColor[2]);
			file.read(parts[i].material.alpha);
			file.read(parts[i].material.shininess);
			file.read(parts[i].material.specularColor[0]);
			file.read(parts[i].material.specularColor[1]);
			file.read(parts[i].material.specularColor[2]);
			file.read(parts[i].material.ambientColor[0]);
			file.read(parts[i].material.ambientColor[1]);
			file.read(parts[i].material.ambientColor[2]);
			file.read(parts[i].toonIndex);
			file.read(parts[i].edgeFlag);

			file.read(parts[i].vertexCount);

			// インデックスバッファを作成
			D3D11_BUFFER_DESC indexBufferDesc = D3D11_BUFFER_DESC(ushort.sizeof * parts[i].vertexCount, D3D11_USAGE.D3D11_USAGE_DEFAULT, D3D11_BIND_FLAG.D3D11_BIND_INDEX_BUFFER, 0, 0);
			D3D11_SUBRESOURCE_DATA indexSubsourceData = D3D11_SUBRESOURCE_DATA(&indices[vertexIndex]);
			if (FAILED(device.CreateBuffer(&indexBufferDesc, &indexSubsourceData, &parts[i].indexBuffer)))
			{
				cleanup();
				throw new Error("Failed create index buffer");
			}

			vertexIndex += parts[i].vertexCount;


			// ファイル名をロード
			ubyte[20] fileName;
			file.read(fileName);

			uint nameLength;
			for (; nameLength < 20; ++nameLength)
			{
				if (fileName[nameLength] == 0 || fileName[nameLength] == 0x2a)
					break;
			}

			if (nameLength > 0)
			{
				// テクスチャファイルパスを取得
				string textureFilePath = std.path.join(directory, cast(string)fileName[0 .. nameLength].idup);

				// テクスチャ読み出しインターフェースをセット
				Bitmap bitmap;
				if (endsWith(textureFilePath, "png"))
					bitmap = new PortableNetworkGraphicsFile(textureFilePath);
				else if (endsWith(textureFilePath, "tga"))
					bitmap = new TruevisionGraphicsAdapterFile(textureFilePath);
				else if (endsWith(textureFilePath, "bmp"))
					bitmap = new MicrosoftWindowsBitmapImageFile(textureFilePath);

				if (bitmap !is null && bitmap.open())
				{
					// テクスチャを読み出し
					uint[] pixelData = bitmap.readAsTrueColor();
					bitmap.close();

					if (pixelData !is null)
					{
						// 作成するテクスチャの設定
						D3D11_TEXTURE2D_DESC textureDesc;
						textureDesc.Width = bitmap.width;
						textureDesc.Height = bitmap.height;
						textureDesc.MipLevels = 1;
						textureDesc.ArraySize = 1;
						textureDesc.Format = DXGI_FORMAT.DXGI_FORMAT_R8G8B8A8_UNORM;
						textureDesc.SampleDesc.Count = 1;
						textureDesc.SampleDesc.Quality = 0;
						textureDesc.Usage = D3D11_USAGE.D3D11_USAGE_IMMUTABLE;
						textureDesc.BindFlags = D3D11_BIND_FLAG.D3D11_BIND_SHADER_RESOURCE;
						textureDesc.CPUAccessFlags = 0;

						// 初期化データとしてピクセルデータを設定
						D3D11_SUBRESOURCE_DATA textureSubsourceData = D3D11_SUBRESOURCE_DATA(cast(const(void)*)pixelData.ptr, bitmap.width * 4);

						// 2Dテクスチャを作成
						ID3D11Texture2D texture;
						if (!FAILED(device.CreateTexture2D(&textureDesc, &textureSubsourceData, &texture)))
						{
							// 2Dテクスチャをシェーダリソースビューに関連づけ
							if (FAILED(device.CreateShaderResourceView(texture, null, &parts[i].textureView)))
								parts[i].textureView = null;

							texture.Release();
						}
					}
				}
			}
		}
	}

	/++
	 + デストラクタ
	 +/
	~this()
	{
		cleanup();
	}

private:

	void cleanup()
	{
		for (uint i; i < parts.length; ++i)
		{
			if (parts[i].textureView !is null)
			{
				parts[i].textureView.Release();
				parts[i].textureView = null;
			}

			if (parts[i].indexBuffer !is null)
			{
				parts[i].indexBuffer.Release();
				parts[i].indexBuffer = null;
			}
		}

		if (vertexBuffer !is null)
		{
			vertexBuffer.Release();
			vertexBuffer = null;
		}
	}
}
